﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Configuration;
using System.Drawing;
using System.IO;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Windows.Forms;
using System.Xml;

using iTextSharp.text.pdf;
using iTextSharp.text.pdf.security;
using System.Data;

public partial class frmInjectData : Form
{
    private static String _folderFormulariPath = Path.GetDirectoryName(Application.ExecutablePath) + "\\Files\\Formulari\\";
    private static String _fileFormulariPath = "TGEN0001-TGEN0001.pdf";

    private static String _folderCarregaPath = Path.GetDirectoryName(Application.ExecutablePath) + "\\Files\\Carrega\\";
    private static String _fileXfaInsertarPathDefault = "xfa_insertar.txt";

    private static String _folderAnnexPath = Path.GetDirectoryName(Application.ExecutablePath) + "\\Files\\Annex";

    private static String _folderCertPath = Path.GetDirectoryName(Application.ExecutablePath) + "\\Files\\Cert\\";
    private static String _fileCertPathDefault = "cda-1_vigente.pfx";


    [STAThread]
    private static void Main()
    {
        Application.EnableVisualStyles();
        Application.SetCompatibleTextRenderingDefault(false);
        Application.Run(new frmInjectData());
    }

    public frmInjectData()
    {
        InitializeComponent();

        // Inicialitzem controls
        txtFile.Text = _folderFormulariPath + _fileFormulariPath;
        txtFolderAnnex.Text = _folderAnnexPath;
        txtFileSign.Text = _folderCertPath + _fileCertPathDefault;

        txtInsert.Text = File.ReadAllText(_folderCarregaPath + _fileXfaInsertarPathDefault);
    }

    private void doExecute()
    {
        try
        {
            // Obtenim array de bytes del pdf a tractar
            byte[] bPDF = File.ReadAllBytes(txtFile.Text);

            // Injectem dades als camps del formulari
            bPDF = InsertData(bPDF);

            // Afegim fitxers annexos
            bPDF = AddAttachment(bPDF);

            // Signem el formulari PDF
            bPDF = SignPDF(bPDF);

            // Escribim el fitxer físic PDF a partir del array de bytes
            File.WriteAllBytes(txtFile.Text.Replace(".pdf", "_omplert.pdf"), bPDF);

            MessageBox.Show("PDF creat correctament!", "", MessageBoxButtons.OK, MessageBoxIcon.Information);
        }
        catch (Exception e)
        {
            MessageBox.Show("Error: " + e.Message + "\r\n\r\n" + e.StackTrace, "", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }
    }

    #region operacions principals

    private byte[] InsertData(byte[] bPDF)
    {
        PdfReader reader = new PdfReader(bPDF);
        using (MemoryStream ms = new MemoryStream())
        {
            PdfStamper stamper = new PdfStamper(reader, ms, '\0', true);
            AcroFields form = stamper.AcroFields;
            XfaForm xfa = form.Xfa;
            XmlNode a = xfa.DatasetsNode;

            XmlNamespaceManager ns = new XmlNamespaceManager(a.OwnerDocument.NameTable);
            ns.AddNamespace("xfa", "http://www.xfa.org/schema/xfa-data/1.0/");

            XmlNode hit = a.SelectSingleNode("//xfa:datasets", ns);
            
            if (hit != null && hit.NodeType == XmlNodeType.Element)
            {
                while (hit.HasChildNodes)
                    hit.RemoveChild(hit.FirstChild);

                int misPosicioInici = txtInsert.Text.IndexOf("<Formulari ");
                int misPosicioFi = txtInsert.Text.IndexOf("</Formulari>");
                string mis = txtInsert.Text.Substring(misPosicioInici, misPosicioFi- misPosicioInici + 12);
                string xfaSenseMis = txtInsert.Text.Remove(misPosicioInici, misPosicioFi - misPosicioInici + 12);

                XmlDocument xmlCustomValues = new XmlDocument();
                xmlCustomValues.LoadXml(xfaSenseMis);

                XmlNode nodeMis = xmlCustomValues.SelectSingleNode("//SUB_MIS/TXT_MIS");
                if (nodeMis != null && nodeMis.NodeType == XmlNodeType.Element)
                    nodeMis.InnerText = mis;

                XmlNode nodeMisCopia = xmlCustomValues.SelectSingleNode("//SUB_MIS/TXT_MIS_COPIA");
                if (nodeMisCopia != null && nodeMisCopia.NodeType == XmlNodeType.Element)
                    nodeMisCopia.InnerText = mis;

                hit.AppendChild(a.OwnerDocument.ImportNode(xmlCustomValues.DocumentElement, true));
            }

            xfa.Changed = true;
            stamper.Close();
            reader.Close();

            return ms.ToArray();
        }
    }

    private byte[] AddAttachment(byte[] bPDF)
    {
        PdfReader reader = new PdfReader(bPDF);
        using (MemoryStream ms = new MemoryStream())
        {
            PdfStamper stamper = new PdfStamper(reader, ms, '\0', true);

            DirectoryInfo d = new DirectoryInfo(txtFolderAnnex.Text);
            FileInfo[] Files = d.GetFiles();
            foreach (FileInfo file in Files)
            {
                PdfFileSpecification fs = PdfFileSpecification.FileEmbedded(stamper.Writer, file.FullName, file.Name, null);
                stamper.AddFileAttachment(file.Name, fs);
            }
            
            stamper.Close();
            reader.Close();

            return ms.ToArray();
        }
    }

    private byte[] SignPDF(byte[] bPDF)
    {
        // Load the certificate into an X509Certificate object.
        X509Certificate2 x509 = new X509Certificate2(File.ReadAllBytes(txtFileSign.Text), "firm@");

        // Output chain information of the selected certificate.
        Org.BouncyCastle.X509.X509CertificateParser cp = new Org.BouncyCastle.X509.X509CertificateParser();
        Org.BouncyCastle.X509.X509Certificate[] chain = new Org.BouncyCastle.X509.X509Certificate[] {
            cp.ReadCertificate(x509.RawData)
        };

        // Creating the reader and the stamper
        PdfReader reader = new PdfReader(bPDF);
        using (MemoryStream ms = new MemoryStream())
        {
            PdfStamper stamper = PdfStamper.CreateSignature(reader, ms, '\0', null, true);

            // Creating the appearance
            PdfSignatureAppearance appearance = stamper.SignatureAppearance;
            appearance.Reason = "Test 1";
            appearance.Location = "Ghent";
            appearance.SetVisibleSignature(new iTextSharp.text.Rectangle(36, 748, 144, 780), 1, "sig");

            // Creating the signature
            IExternalSignature externalSignature = new X509Certificate2Signature(x509, DigestAlgorithms.SHA1);
            MakeSignature.SignDetached(appearance, externalSignature, chain, null, null, null, 0, CryptoStandard.CADES);

            stamper.Close();
            reader.Close();

            return ms.ToArray();
        }
    }

    #endregion

    #region button events

    private void btnFile_Click(object sender, EventArgs e)
    {
        OpenFileDialog openFileDialog = new OpenFileDialog();
        openFileDialog.Filter = "PDF files (*.pdf)|*.pdf|All files (*.*)|*.*";
        openFileDialog.InitialDirectory = _folderFormulariPath;
        openFileDialog.Title = "Open PDF file";
        openFileDialog.Multiselect = false;
        openFileDialog.CheckFileExists = true;

        if (openFileDialog.ShowDialog() == DialogResult.OK)
            this.txtFile.Text = openFileDialog.FileName;
    }

    private void btnFolderAnnex_Click(object sender, EventArgs e)
    {
        using (FolderBrowserDialog dlg = new FolderBrowserDialog())
        {
            dlg.Description = "Select a folder";
            dlg.SelectedPath = _folderAnnexPath;

            if (dlg.ShowDialog() == DialogResult.OK)
                txtFolderAnnex.Text = dlg.SelectedPath;
        }
    }

    private void btnFileSign_Click(object sender, EventArgs e)
    {
        OpenFileDialog openFileDialog = new OpenFileDialog();
        openFileDialog.Filter = "All files (*.*)|*.*";
        openFileDialog.InitialDirectory = _folderCertPath;
        openFileDialog.Title = "Select a certificate";
        openFileDialog.Multiselect = false;
        openFileDialog.CheckFileExists = true;

        if (openFileDialog.ShowDialog() == DialogResult.OK)
            txtFileSign.Text = openFileDialog.FileName;
    }

    private void btnExecutar_Click(object sender, EventArgs e)
    {
        btnExecutar.Enabled = false;

        if (!String.IsNullOrEmpty(txtFile.Text) && !String.IsNullOrEmpty(txtFolderAnnex.Text) && !String.IsNullOrEmpty(txtFileSign.Text))
            doExecute();
        else
            MessageBox.Show("Cal omplir tots els imputs", "", MessageBoxButtons.OK, MessageBoxIcon.Warning);

        btnExecutar.Enabled = true;
    }

    #endregion

}
